from jinja2 import Template, StrictUndefined, Environment
from pandas import Timestamp
import typing as t


class SqlQuery:
    def __init__(self, template_string: str, **kwargs):
        env = Environment(undefined=StrictUndefined)
        self.template: Template = env.from_string(template_string)
        self.bindings = kwargs

    def add_bindings(self, *args: t.Any, **kwargs: t.Any):
        self.bindings = self.bindings | dict(*args, **kwargs)

    def render(self, *args: t.Any, **kwargs: t.Any) -> str:
        bindings_original = self.bindings | dict(*args, **kwargs)

        bindings_curated = {}
        for key, value in bindings_original.items():
            if isinstance(value, SqlQuery):
                bindings_curated[key] = f"({value.render()})"
            elif isinstance(value, list):
                values_list = [
                    f"'{x}'" if isinstance(x, str) else str(x) for x in value
                ]
                bindings_curated[key] = f"({','.join(values_list)})"
            elif isinstance(value, str):
                try:
                    value_dt = Timestamp(value)
                    bindings_curated[key] = f"'{str(value_dt)}'"
                except ValueError:
                    bindings_curated[key] = f"'{value}'"
            elif isinstance(value, SqlExpr):
                bindings_curated[key] = value.value
            else:
                bindings_curated[key] = value

        return self.template.render(**bindings_curated)

class SqlExpr():
    def __init__(self, value: str, identifier: str = "`", add_identifier: bool = True):
        if add_identifier:
            self.value = f"{identifier}{value}{identifier}"
        else:
            self.value = value



import dagster as dg
from typing import Sequence, Type
from dagster._core.storage.db_io_manager import TableSlice
from duckdb import BinderException, DuckDBPyConnection, IOException
from dagster_duckdb.io_manager import DuckDbClient
from dagster._utils.backoff import backoff

from dagster_db.helpers.generic_db import table_slice_to_schema_table
from dagster_db.helpers.duckdb import execute_duckdb
from dagster_db.helpers.sql_query import get_sample_md, get_table_schema, get_rows, glimpse
from dagster_db.type_handlers.custom_type_handler import CustomDbTypeHandler
from dagster_db.query.sql_query import SqlQuery


class DuckDbQueryTypeHandler(CustomDbTypeHandler[SqlQuery, DuckDBPyConnection]):
    @property
    def supported_types(self) -> Sequence[Type[object]]:
        return [SqlQuery]

    def validate_obj_db(
        self,
        context,
        obj_db: SqlQuery,
        connection: DuckDBPyConnection,
    ):
        return

    def db_safe_transformations(
        self,
        context,
        obj: SqlQuery,
        connection: DuckDBPyConnection,
    ) -> SqlQuery:
        return obj

    def output_metadata(
        self,
        context: dg.OutputContext,
        obj: SqlQuery,
        obj_db: SqlQuery,
        connection: DuckDBPyConnection,
    ):
        return {
            "sample_obj": dg.MarkdownMetadataValue(get_sample_md(obj, connection)),
            "sample_obj_db": dg.MarkdownMetadataValue(get_sample_md(obj_db, connection)),
            "rows": dg.IntMetadataValue(get_rows(obj, connection)),
            "table_schema": dg.TableSchemaMetadataValue(get_table_schema(obj, connection)),
        }

    def _load_into_db(
        self,
        table_schema,
        obj: SqlQuery,
        connection: DuckDBPyConnection,
    ):
        ctas_query = SqlQuery("CREATE TABLE IF NOT EXISTS {{ table_schema }} AS SELECT * FROM {{ obj }}", table_schema=table_schema, obj=obj)
        execute_duckdb(ctas_query, connection)
        if not connection.fetchall():
            insert_query = SqlQuery("INSERT INTO {{ table_schema }} SELECT * FROM {{ obj }}", table_schema=table_schema, obj=obj)
            execute_duckdb(insert_query, connection)

    def handle_output(
        self,
        context: dg.OutputContext,
        table_slice: TableSlice,
        obj: SqlQuery,
        connection: DuckDBPyConnection,
    ):
        table_schema = table_slice_to_schema_table(table_slice)
        create_schema_query = SqlQuery("CREATE SCHEMA IF NOT EXISTS {{ schema }}", schema=table_slice.schema)
        execute_duckdb(create_schema_query, connection)

        try:
            backoff(
                self._load_into_db,
                retry_on=(IOException,),
                kwargs={
                    "table_schema": table_schema,
                    "obj": obj,
                    "connection": connection,
                },
                max_retries=1,
            )
        except BinderException:
            obj_existing = self.load_input(context, table_slice, connection)
            msg = f"""`obj` incompatible with existing table

            `obj`
            {glimpse(obj, connection)}

            `obj_existing`
            {glimpse(obj_existing, connection)}
            """
            context.log.error(msg)

        return

    def load_input(
        self,
        context: dg.InputContext | dg.OutputContext,
        table_slice: TableSlice,
        connection: DuckDBPyConnection,
    ) -> SqlQuery:
        if table_slice.partition_dimensions and len(context.asset_partition_keys) == 0:
            raise ValueError(
                f"{table_slice.partition_dimensions=} incompatible with {context.asset_partition_keys=}"
            )

        query = DuckDbClient.get_select_statement(table_slice)
        return SqlQuery(query)



from typing import Optional
import pandas as pd
import polars as pl
import dagster as dg
from duckdb import DuckDBPyConnection
from dagster_db.helpers.duckdb import execute_duckdb
from dagster_db.query.sql_query import SqlQuery


def get_sample_md(obj: SqlQuery, connection: DuckDBPyConnection, n: int = 10) -> Optional[str]:
    sample_query = SqlQuery("SELECT * FROM {{ obj }} ORDER BY RAND() LIMIT {{ n }}", obj=obj, n=n)
    df = execute_duckdb(sample_query, connection, return_type=pd.DataFrame)
    return df.to_markdown()

def get_table_schema(obj: SqlQuery, connection: DuckDBPyConnection) -> dg.TableSchema:
    sample_query = SqlQuery("SELECT * FROM {{ obj }} LIMIT 0", obj=obj)
    result = execute_duckdb(sample_query, connection, return_type=DuckDBPyConnection)
    description = result.description
    assert description is not None
    return dg.TableSchema(
        columns=[
            dg.TableColumn(name=field[0], type=field[1])
            for field in description
        ]
    )

def get_rows(obj: SqlQuery, connection: DuckDBPyConnection) -> int:
    count_query = SqlQuery("SELECT COUNT(*) as count FROM {{ obj }}", obj=obj)
    result = execute_duckdb(count_query, connection, return_type=tuple)
    return result[0]

def glimpse(obj: SqlQuery, connection: DuckDBPyConnection, n: int = 10) -> str:
    sample_query = SqlQuery("SELECT * FROM {{ obj }} LIMIT {{ n }}", obj=obj, n=n)
    df = execute_duckdb(sample_query, connection, return_type=pl.DataFrame)
    return df.glimpse(return_as_string=True)



from typing import Optional
import polars as pl
import numpy as np
import dagster as dg

def get_sample_md(obj: pl.DataFrame, n_max=10) -> Optional[str]:
    n = min(n_max, obj.height)
    return obj.sample(n, with_replacement=False).to_pandas().to_markdown()


def get_summary_md(obj: pl.DataFrame) -> Optional[str]:
    try:
        return (
            obj.select(
                [
                    key
                    for key, value in obj.schema.items()
                    if value not in [pl.Utf8, pl.Boolean]
                ]
            )
            .describe()
            .to_pandas()
            .set_index("statistic")
            .fillna(np.inf)
            .to_markdown()
        )
    except TypeError:
        # No numeric columns
        return "No numeric columns to summarise"


def get_table_schema(
    obj: pl.DataFrame
) -> dg.TableSchema:
    return dg.TableSchema(
        columns=[
            dg.TableColumn(name=name, type=str(dtype))
            for name, dtype in zip(obj.columns, obj.dtypes)
        ]
    )



from typing import (
    Optional,
    Sequence,
    Type,
    Any,
)

from dagster._core.storage.db_io_manager import DbIOManager, DbClient
import dagster as dg

from dagster_db.type_handlers.custom_type_handler import CustomDbTypeHandler


class CustomDbIOManager(DbIOManager):
    def __init__(
        self,
        *,
        type_handlers: Sequence[CustomDbTypeHandler],
        db_client: DbClient,
        database: str,
        schema: Optional[str] = None,
        io_manager_name: Optional[str] = None,
        default_load_type: Optional[Type] = None,
    ):
        # TODO: validate type_handlers
        # TODO: don't inherit init?

        super().__init__(
            type_handlers=type_handlers,
            db_client=db_client,
            database=database,
            schema=schema,
            io_manager_name=io_manager_name,
            default_load_type=default_load_type,
        )

        self._handlers_by_type: dict[type[Any], CustomDbTypeHandler] # type: ignore

    def handle_output(self, context: dg.OutputContext, obj: object) -> None:
        obj_type = type(obj)
        self._check_supported_type(obj_type)
        table_slice = self._get_table_slice(context, context)

        handler = self._handlers_by_type[obj_type]
        with self._db_client.connect(context, table_slice) as conn:
            obj_db = handler.db_safe_transformations(context, obj, conn)
            handler.validate_obj_db(context, obj_db, conn)

            context.log.debug("all validation successful")
            super().handle_output(context, obj)
            context.add_output_metadata(handler.output_metadata(context, obj, obj_db, conn))

    def load_input(self, context: dg.InputContext) -> object:
        return super().load_input(context)


def build_custom_duckdb_io_manager(
    type_handlers: Sequence[CustomDbTypeHandler],
    default_load_type: Optional[Type] = None,
    io_manager_name: str = "DuckDbIoManager"
) -> dg.IOManagerDefinition:
    @dg.io_manager(config_schema=DuckDBIOManager.to_config_schema())
    def duckdb_io_manager(init_context):
        return CustomDbIOManager(
            type_handlers=type_handlers,
            db_client=DuckDbClient(),
            io_manager_name=io_manager_name,
            database=init_context.resource_config["database"],
            schema=init_context.resource_config.get("schema"),
            default_load_type=default_load_type,
        )

    return duckdb_io_manager



import pandas as pd
import polars as pl
from duckdb import DuckDBPyConnection
from dagster_db.query.sql_query import SqlQuery

from typing import Type, TypeVar

T = TypeVar("T", list, tuple, pd.DataFrame, pl.DataFrame, DuckDBPyConnection)

def execute_duckdb(query: SqlQuery, connection: DuckDBPyConnection, return_type: Type[T] = DuckDBPyConnection) -> T:
    result = connection.execute(query.render())
    if return_type is list:
        return_value = result.fetchall()
    elif return_type is tuple:
        return_value = result.fetchone()
    elif return_type is pd.DataFrame:
        return_value = result.df()
    elif return_type is pl.DataFrame:
        return_value = result.pl()
    elif return_type is DuckDBPyConnection:
        return_value = result
    else:
        raise TypeError(f"Unsupported {return_type=}")

    return return_value # type: ignore

